
/*
 * This is a serial library.
 *
 * History:
 * Date		Author			Comment
 * 06-02-2006	Victor Yu.		Create it.
 */
#include	<stdio.h>
#include	<stdlib.h>
#include	<fcntl.h>
#include	<errno.h>
#include	<string.h>
#include	<unistd.h>
#include	<pthread.h>
#include	<sys/select.h>
#include	<asm/ioctls.h>
#include	"sio.h"

#ifndef CMSPAR
#define CMSPAR			010000000000
#endif

#define MAX_SPEED		921600

#ifndef MOXA_SET_OP_MODE
#define MOXA_SET_OP_MODE	(0x400+66)
#endif
#ifndef MOXA_GET_OP_MODE
#define MOXA_GET_OP_MODE	(0x400+67)
#endif

struct speed_table_struct {
	int	index;
	int	speed;
};
static struct speed_table_struct	speed_table[]={
	{B50, 50},
	{B75, 75},
	{B110, 110},
	{B134, 134},
	{B150, 150},
	{B300, 300},
	{B600, 600},
	{B1200, 1200},
	{B1800, 1800},
	{B2400, 2400},
	{B4800, 4800},
	{B7200, 7200},
	{B9600, 9600},
	{B19200, 19200},
	{B38400, 38400},
	{B57600, 57600},
	{B115200, 115200},
	{B230400, 230400},
	{B460800, 460800},
	{B921600, 921600}
};
#define SPEED_TABLE_NO	(sizeof(speed_table)/sizeof(struct speed_table_struct))

typedef struct _serial_info_struct {
	int				fd;
	int				port;
	int				interface;
	struct termios			termios;
	pthread_mutex_t			pmutex;
	struct _serial_info_struct	*next;
} serial_info;

static serial_info	*serial_info_head=NULL, *serial_info_tail=NULL;
static pthread_mutex_t	smutex=PTHREAD_MUTEX_INITIALIZER;
static int		serial_init_flag=0;

static serial_info 	*searchport(int port)
{
	serial_info	*now;

	now = serial_info_head;
	while ( now ) {
		if ( now->port == port )
			break;
		now = now->next;
	}
	return now;
}

#define GET_PORT(ser, no) {	\
	pthread_mutex_lock(&smutex);	\
	if ( serial_init_flag == 0 ) {	\
		pthread_mutex_unlock(&smutex);	\
		return SIO_NOINIT;	\
	}	\
	(ser) = searchport((no)); \
	pthread_mutex_unlock(&smutex); \
	if ( (ser) == NULL )	\
		return SIO_BADPORT;	\
}

int	sio_port_to_fd(int port)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	return serial->fd;
}

int	sio_fd_to_port(int fd)
{
	serial_info	*now;

	pthread_mutex_lock(&smutex);
	now = serial_info_head;
	while ( now ) {
		if ( now->fd == fd )
			break;
		now = now->next;
	}
	pthread_mutex_unlock(&smutex);
	return now->port;
}

int	sio_init(void)
{
	pthread_mutex_lock(&smutex);
	if ( serial_init_flag ) {
		pthread_mutex_unlock(&smutex);
		return SIO_OK;
	}
	serial_init_flag = 1;
	pthread_mutex_unlock(&smutex);
	return SIO_OK;
}

void 	sio_release(void)
{
	serial_info	*now;

	pthread_mutex_lock(&smutex);
	if ( serial_init_flag == 0 ) {
		pthread_mutex_unlock(&smutex);
		return;
	}
	now = serial_info_head;
	while ( now ) {
		close(now->fd);
		serial_info_head = now->next;
		free(now);
		now = serial_info_head;
	}
	serial_init_flag = 0;
	pthread_mutex_unlock(&smutex);
}

int	sio_open(int port)
{
	serial_info	*serial;
	char		ttyname[64];

#if 1	// add by Victor Yu. 06-16-2006
	sio_init();
	pthread_mutex_lock(&smutex);
#else
	pthread_mutex_lock(&smutex);
	if ( serial_init_flag == 0 ) {
		pthread_mutex_unlock(&smutex);
		return SIO_NOINIT;
	}
#endif
	serial = searchport(port);
	if ( serial ) {
		pthread_mutex_unlock(&smutex);
		return SIO_OPENFAIL;
	}
	serial = malloc(sizeof(serial_info));
	if ( serial == NULL ) {
		pthread_mutex_unlock(&smutex);
		return SIO_FAIL;
	}
	memset(serial, 0, sizeof(serial_info));
	serial->port = port;
	pthread_mutex_init(&serial->pmutex, NULL);
	sprintf(ttyname, "/dev/ttyM%d", port);
	serial->fd = open(ttyname, O_RDWR|O_NONBLOCK);
	if ( serial->fd <= 0 ) {
		free(serial);
		pthread_mutex_unlock(&smutex);
		return SIO_OPENFAIL;
	}
	tcgetattr(serial->fd, &serial->termios);
	serial->termios.c_lflag = 0;
	serial->termios.c_oflag = 0;
	serial->termios.c_iflag = 0;
	serial->termios.c_cflag = B38400 | CS8 | CLOCAL | CREAD;
	serial->termios.c_cc[VTIME] = 0;
	serial->termios.c_cc[VMIN] = 0;
	tcsetattr(serial->fd, TCSANOW, &serial->termios);
	tcflush(serial->fd, TCIOFLUSH);
	serial->interface = RS232_MODE;
	ioctl(serial->fd, MOXA_SET_OP_MODE, &serial->interface);
	if ( serial_info_head == NULL )
		serial_info_head = serial;
	else
		serial_info_tail->next = serial;
	serial_info_tail = serial;
	pthread_mutex_unlock(&smutex);
	return SIO_OK;
}

int	sio_close(int port)
{
	serial_info	*now, *priv;

	pthread_mutex_lock(&smutex);
	now = priv = serial_info_head;
	while ( now ) {
		if ( now->port == port ) {
			if ( now == serial_info_head ) {
				if ( serial_info_tail == serial_info_head )
					serial_info_tail = NULL;
				serial_info_head = serial_info_head->next;
			} else if ( now == serial_info_tail )
				serial_info_tail = priv;
			priv->next = now->next;
			close(now->fd);
			free(now);
			break;
		}
		priv = now;
		now = now->next;
	}
	pthread_mutex_unlock(&smutex);
	return SIO_OK;
}

int	sio_read(int port, char *buf, int len)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	if ( buf == NULL || len < 0 )
		return SIO_BADPARM;
	if ( len == 0 )
		return 0;
	return read(serial->fd, buf, len);
}

int	sio_block_read(int port, char *buf, int len)
{
	serial_info	*serial;
	int		fd, ret=0;
	struct timeval	time;
	fd_set		readfds;

	GET_PORT(serial, port);
	if ( buf == NULL || len < 0 )
		return SIO_BADPARM;
	if ( len == 0 )
		return 0;
	fd = sio_port_to_fd(port);
	while ( ret < len ) {
		time.tv_usec = 10000;	// 10msec
		time.tv_sec = 0;
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		if ( select(fd+1, &readfds, NULL, NULL, &time) >= 0 ) {
			if ( FD_ISSET(fd, &readfds) ) {
				ret += read(fd, &buf[ret], len-ret);
			}
		}
	}
	return ret;
}

int	sio_write(int port, char *buf, int len)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	if ( buf == NULL || len < 0 )
		return SIO_BADPARM;
	if ( len == 0 )
		return 0;
	return write(serial->fd, buf, len);
}

int	sio_block_write(int port, char *buf, int len)
{
	serial_info	*serial;
	int		fd, ret=0;
	struct timeval	time;
	fd_set		writefds;

	GET_PORT(serial, port);
	if ( buf == NULL || len < 0 )
		return SIO_BADPARM;
	if ( len == 0 )
		return 0;
	fd = sio_port_to_fd(port);
	while ( ret < len ) {
		time.tv_usec = 10000;	// 10msec
		time.tv_sec = 0;
		FD_ZERO(&writefds);
		FD_SET(fd, &writefds);
		if ( select(fd+1, NULL, &writefds, NULL, &time) >= 0 ) {
			if ( FD_ISSET(fd, &writefds) ) {
				ret += write(fd, &buf[ret], len-ret);
			}
		}
	}
	while ( 1 ) {
		time.tv_usec = 10000;	// 10msec
		time.tv_sec = 0;
		FD_ZERO(&writefds);
		FD_SET(0, &writefds);
		select(1, NULL, NULL, NULL, &time);
		if ( sio_oqueue(port) == 0 )
			break;
	}
	return ret;
}

int	sio_getch(int port)
{
	serial_info	*serial;
	int		ret;

	GET_PORT(serial, port);
	if ( read(serial->fd, (char *)&ret, 1) == 0 )
		return SIO_NODATA;
	return ret;
}

int	sio_putch(int port, int term)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	if ( write(serial->fd, (char *)&term, 1) == 0 )
		return SIO_NODATA;
	return SIO_OK;
}

int 	sio_linput(int port, char *buf, int len, int term)
{
	serial_info	*serial;
	int		i;

	GET_PORT(serial, port);
	if ( buf == NULL || len < 0 )
		return SIO_BADPARM;
	i = len;
	while ( len ) {
		read(serial->fd, &buf[i-len], 1);
		if ( buf[i-len] == (char)term ) {
			len--;
			break;
		}
		len--;
	}
	return (i-len);
}

int	sio_ioctl(int port, int baud, int mode)
{
	serial_info	*serial;
	int		i;

	GET_PORT(serial, port);
	for ( i=0; i<SPEED_TABLE_NO && speed_table[i].speed != baud; i++ );
	if ( i == SPEED_TABLE_NO )
		return SIO_BADPARM;
	pthread_mutex_lock(&serial->pmutex);
	serial->termios.c_cflag &= ~(CBAUD|CSIZE|CSTOPB|PARENB|PARODD|CMSPAR);
	serial->termios.c_cflag |= speed_table[i].index;
	switch ( mode & BIT_MASK ) {
	case BIT_5 :
		serial->termios.c_cflag |= CS5;
		break;
	case BIT_6 :
		serial->termios.c_cflag |= CS6;
		break;
	case BIT_7 :
		serial->termios.c_cflag |= CS7;
		break;
	case BIT_8 :
		serial->termios.c_cflag |= CS8;
		break;
	}
	if ( mode & STOP_2 )
		serial->termios.c_cflag |= CSTOPB;
	switch ( mode & P_MASK ) {
	case P_EVEN :
		serial->termios.c_cflag |= PARENB;
		break;
	case P_ODD :
		serial->termios.c_cflag |= (PARENB|PARODD);
		break;
	case P_SPC :
		serial->termios.c_cflag |= (PARENB|CMSPAR);
		break;
	case P_MRK :
		serial->termios.c_cflag |= (PARENB|CMSPAR|PARODD);
		break;
	case P_NONE :
		break;
	}
	i = tcsetattr(serial->fd, TCSANOW, &serial->termios);
	pthread_mutex_unlock(&serial->pmutex);
	if ( i )
		return SIO_FAIL;
	return SIO_OK;
}

int 	sio_flush(int port, int func)
{
	serial_info	*serial;
	int		ret;

	GET_PORT(serial, port);
	if ( func != FLUSH_INPUT && func != FLUSH_OUTPUT && func != FLUSH_INOUT )
		return SIO_BADPARM;
	switch ( func ) {
	case FLUSH_INPUT :
		ret = tcflush(serial->fd, TCIFLUSH);
		break;
	case FLUSH_OUTPUT :
		ret = tcflush(serial->fd, TCOFLUSH);
		break;
	case FLUSH_INOUT :
		ret = tcflush(serial->fd, TCIOFLUSH);
		break;
	}
	if ( ret )
		return SIO_FAIL;
	return SIO_OK;
}

long	sio_iqueue(int port)
{
	serial_info	*serial;
	long		ret;

	GET_PORT(serial, port);
	if ( ioctl(serial->fd, TIOCINQ, &ret) )
		return SIO_FAIL;
	return ret;
}

long	sio_oqueue(int port)
{
	serial_info	*serial;
	long		ret;

	GET_PORT(serial, port);
	if ( ioctl(serial->fd, TIOCOUTQ, &ret) )
		return SIO_FAIL;
	return ret;
}

int	sio_lstatus(int port)
{
	serial_info	*serial;
	int		ret=0, status;

	GET_PORT(serial, port);
	if ( ioctl(serial->fd, TIOCMGET, &status) )
		return SIO_FAIL;
	if ( status & TIOCM_CAR )
		ret |= S_CD;
	if ( status & TIOCM_RNG )
		ret |= S_RI;
	if ( status & TIOCM_DSR )
		ret |= S_DSR;
	if ( status & TIOCM_CTS )
		ret |= S_CTS;
	return ret;
}

int	sio_lctrl(int port, int mode)
{
	serial_info	*serial;
	int		ctrl=0;

	GET_PORT(serial, port);
	if ( mode & C_RTS ) {
		if ( serial->termios.c_cflag & CRTSCTS )
			return SIO_RTS_BY_HW;
		ctrl |= TIOCM_RTS;
	}
	if ( mode & C_DTR )
		ctrl |= TIOCM_DTR;
	if ( ioctl(serial->fd, TIOCMSET, &ctrl) )
		return SIO_FAIL;
	return SIO_OK;
}

int	sio_break_ex(int port, int time)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	if ( time < 0 )
		return SIO_BADPARM;
	if ( time == 0 )
		return SIO_OK;
	ioctl(serial->fd, TIOCSBRK, NULL);
	usleep(time/1000);
	ioctl(serial->fd, TIOCCBRK, NULL);
	return SIO_OK;
}

int	sio_flowctrl(int port, int mode)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	pthread_mutex_lock(&serial->pmutex);
	serial->termios.c_cflag &= ~CRTSCTS;
	serial->termios.c_iflag &= ~(IXON|IXOFF);
	if ( mode & HW_FLOWCTRL )
		serial->termios.c_cflag |= CRTSCTS;
	if ( mode & SW_FLOWCTRL )
		serial->termios.c_iflag |= (IXON|IXOFF);
	tcsetattr(serial->fd, TCSANOW, &serial->termios);
	pthread_mutex_unlock(&serial->pmutex);
	return SIO_OK;
}

long	sio_getbaud(int port)
{
	serial_info	*serial;
	int		i, index;

	GET_PORT(serial, port);
	pthread_mutex_lock(&serial->pmutex);
	index = serial->termios.c_cflag & CBAUD;
	pthread_mutex_unlock(&serial->pmutex);
	for ( i=0; i<SPEED_TABLE_NO && index != speed_table[i].index; i++ );
	if ( i == SPEED_TABLE_NO )
		return SIO_FAIL;
	return (long)speed_table[i].speed;
}

int	sio_getmode(int port)
{
	serial_info	*serial;
	int		ret=0;

	GET_PORT(serial, port);
	pthread_mutex_lock(&serial->pmutex);
	switch( serial->termios.c_cflag & CSIZE ) {
	case CS5 :
		ret |= BIT_5;
		break;
	case CS6 :
		ret |= BIT_6;
		break;
	case CS7 :
		ret |= BIT_7;
		break;
	case CS8 :
		ret |= BIT_8;
		break;
	}
	if ( serial->termios.c_cflag & CSTOPB )
		ret |= STOP_2;
	if ( serial->termios.c_cflag & PARENB ) {
		if ( serial->termios.c_cflag & CMSPAR ) {
			if ( serial->termios.c_cflag & PARODD )
				ret |= P_MRK;
			else
				ret |= P_SPC;
		} else {
			if ( serial->termios.c_cflag & PARODD )
				ret |= P_ODD;
			else
				ret |= P_EVEN;
		}
	} else
		ret |= P_NONE;
	pthread_mutex_unlock(&serial->pmutex);
	return ret;
}

int	sio_getflow(int port)
{
	serial_info	*serial;
	int		ret=0;

	GET_PORT(serial, port);
	pthread_mutex_lock(&serial->pmutex);
	if ( serial->termios.c_cflag & CRTSCTS )
		ret |= HW_FLOWCTRL;
	if ( serial->termios.c_iflag & (IXON|IXOFF) )
		ret |= SW_FLOWCTRL;
	pthread_mutex_unlock(&serial->pmutex);
	return ret;
}

int	sio_DTR(int port, int mode)
{
	serial_info	*serial;
	int		i=TIOCM_DTR;

	GET_PORT(serial, port);
	if ( mode != 0 && mode != 1 )
		return SIO_BADPARM;
	if ( mode )
		ioctl(serial->fd, TIOCMBIS, &i);
	else
		ioctl(serial->fd, TIOCMBIC, &i);
	return SIO_OK;
}

int	sio_RTS(int port, int mode)
{
	serial_info	*serial;
	int		i=TIOCM_RTS;

	GET_PORT(serial, port);
	if ( mode != 0 && mode != 1 )
		return SIO_BADPARM;
	if ( serial->termios.c_cflag & CRTSCTS )
		return SIO_RTS_BY_HW;
	if ( mode )
		ioctl(serial->fd, TIOCMBIS, &i);
	else
		ioctl(serial->fd, TIOCMBIC, &i);
	return SIO_OK;
}

int	sio_baud(int port, long speed)
{
	serial_info	*serial;
	int		i;

	GET_PORT(serial, port);
	for ( i=0; i<SPEED_TABLE_NO && speed_table[i].speed != (int)speed; i++ );
	if ( i == SPEED_TABLE_NO )
		return SIO_BADPARM;
	pthread_mutex_lock(&serial->pmutex);
	serial->termios.c_cflag &= ~CBAUD;
	serial->termios.c_cflag |= speed_table[i].index;
	tcsetattr(serial->fd, TCSANOW, &serial->termios);
	pthread_mutex_unlock(&serial->pmutex);
	return SIO_OK;
}

int	sio_set_interface(int port, int mode)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	if ( mode != RS232_MODE && mode != RS485_2WIRE_MODE && mode != RS422_MODE && mode != RS485_4WIRE_MODE )
		return SIO_BADPARM;
	pthread_mutex_lock(&serial->pmutex);
	serial->interface = mode;
	ioctl(serial->fd, MOXA_SET_OP_MODE, &mode);
	pthread_mutex_unlock(&serial->pmutex);
	return SIO_OK;
}

int 	sio_get_interface(int port, int *mode)
{
	serial_info	*serial;

	GET_PORT(serial, port);
	*mode = serial->interface;
	return SIO_OK;
}
